module Compilers.Emacs where

import qualified Data.Set as Set
import qualified Data.List as List

import Syntax.Position
import Syntax.Internal
import Utils.Files
import Utils.Unicode
import Utils.Pairs
import Utils.Lists
import Utils.Functions

emacsFile :: FilePath -> FilePath
emacsFile file =
    case splitFilePath file of
	(path,name,ext)	-> path ++ "." ++ name ++ ext ++ ".el"

escape = concatMap esc
    where
	escchars = "\\\""
	esc c	| c `elem` escchars = ['\\',c]
		| otherwise	    = [c]

-- Constructs a string of the following form:
--
-- (agdaLight-add-type-names '("Maybe" "Nat"))
-- (agdaLight-add-type-operators '("×"))
-- (agdaLight-add-constructor-names '("Just" "zero"))
-- (agdaLight-add-constructor-operators '())
-- (agdaLight-add-axiom-names '("Not" "viewL"))
-- (agdaLight-add-axiom-operators '("=="))
-- (agdaLight-add-function-names '("nil" "queue"))
-- (agdaLight-add-function-operators '(">>="))
--
-- It is then up to some Emacs mode to define the above functions in a
-- suitable manner, and to load the file containing the string.

toEmacs :: Names -> String
toEmacs names =
  compose
    (map makeEntry
      [ ("type", typs)
      , ("constructor", cons)
      , ("axiom", axs)
      , ("function", defs)
      ]
    )
  $ ""
    where
        compose = foldr (.) id

        -- Can be optimised (currently makes two passes).
        makeEntry p = makeEntry' (not . isOperator) "names" p
                      . makeEntry' isOperator "operators" p
        makeEntry' isKind kind (typ, names) =
          showString "(agdaLight-add-"
          . showString typ
          . showString "-"
          . showString kind
          . showString " '("
          . compose (List.intersperse (showChar ' ') $
                     map (quote . nameString) $ filter isKind names)
          . showString "))\n"

        quote s = showChar '"' . showString (escape s) . showChar '"'

	ns = Set.toList names

	typs  = [ x | ITyp _ x <- ns ]
	cons  = [ x | ICon _ x <- ns ]
	defs  = [ x | IDef _ x <- ns ]
	axs   = [ x | IAx  _ x <- ns ]

type Names = Set.Set DefName

data DefName = IDef { range :: Range, getName :: Name }
	     | IAx  { range :: Range, getName :: Name }
	     | ICon { range :: Range, getName :: Name }
	     | ITyp { range :: Range, getName :: Name }

instance Eq DefName where
    x == y  = getName x == getName y

instance Ord DefName where
    compare x y = compare (getName x) (getName y)

definedNames :: Program -> [DefName]
definedNames = concatMap defName
    where
	defName (NormalDef r x _ _ _ _)	    = [IDef r x]
	defName (NewAxiom r x _ _)	    = [IAx r x]
	defName (NewSig r x _ _ _ _ cs)	    = [ITyp r x]
	defName (NewData _ r x _ _ _ cs)    = ITyp r x : map cons cs
	    where
		cons (Constructor r x _ _)  = ICon r x
	defName (Rewrite _ _)		    = []
	defName (Mutual _ ds)		    = concatMap defName ds
	defName (Synonym r x _)		    = [IDef r x]

generateEmacsFile :: FilePath -> Program -> IO ()
generateEmacsFile file p = writeFileUTF8 (emacsFile file) $ toEmacs names
    where
	names = foldr Set.insert Set.empty $ definedNames p
