-- Message "Wrong Constructor"
-- Source :: Normalizing If-Expressions
-- Author :: Andreas Abel
-- Created:: 2004-12-07

-- The booleans -----------------------------------------------------------
data Bool : Set where
  false	: Bool
  true	: Bool

-- If we define if in terms of ifD
-- we get problems with lambdas in FOL terms
if : (A:Set) -> Bool -> A -> A -> A
if A true x y = x
if A false x y = y

sort Prop : Type where
  Pi Set Prop -> Prop

Eq : (A : Set) -> A -> A -> Prop


-- If expressions over variables in some set A

data IfExp (A:Set) : Set where
    Bit : Bool -> IfExp A
    Var : A -> IfExp A
    If  : IfExp A -> IfExp A -> IfExp A -> IfExp A

-- Semantics

eval : (A:Set) -> (A -> Bool) -> (IfExp A -> Bool)
eval A f (Bit b) = b
eval A f (Var a) = f a
eval A f (If c t e) = if Bool (eval A f c) (eval A f t) (eval A f e)

-- Normal if expressions

data NIfExp (A:Set) : Set where
    NBit : Bool -> NIfExp A
    NVar : A -> NIfExp A
    NIf  : A -> NIfExp A -> NIfExp A -> NIfExp A

-- Embedding into IfExp

emb : (A:Set) -> NIfExp A -> IfExp A
emb A (NBit b) = Bit b
emb A (NVar a) = Var a
emb A (NIf a t e) = If (Var a) (emb A t) (emb A e)

-- Normalization of if expressions

-- composition of normal expressions

nif : (A:Set) -> NIfExp A -> NIfExp A -> NIfExp A -> NIfExp A
nif A (NBit true) t e = t
nif A (NBit false) t e = e
nif A (NVar a) t e = NIf a t e
nif A (NIf a b c) t e = NIf a (nif A b t e) (nif A c t e)

-- normalization

norm : (A:Set) -> IfExp A -> NIfExp A
norm A (Bit b) = NBit b
norm A (Var a) = NVar a
norm A (If c t e) = nif A (norm A c) (norm A t) (norm A e)

-- Soundness of normalization ----------------------------------------

-- Induction for IfExp

indIfExp : (A:Set) -> (P : IfExp A -> Prop) ->
            ((b:Bool) -> P (Bit b)) ->
            ((a:A) -> P (Var a)) ->
            ((c,t,e : IfExp A) -> P c -> P t -> P e -> P (If c t e)) ->
            (e : IfExp A) -> P e
indIfExp A P bit var if (Bit b) = bit b
indIfExp A P bit var if (Var a) = var a
indIfExp A P bit var if (If c t e)  = if c t e (indIfExp A P bit var if c)
                                               (indIfExp A P bit var if t)  
                                               (indIfExp A P bit var if e)

-- Induction for NIfExp

indNIfExp : (A:Set) -> (P : NIfExp A -> Prop) ->
             ((b:Bool) -> P (Bit b)) ->
                           -- ^ here is the wrong constructor
             ((a:A) -> P (Var a)) ->
                        -- ^ here is the wrong constructor
             ((a:A) -> (t,e : NIfExp A) -> P t -> P e -> 
                P (If a t e)) ->
                -- ^ here is the wrong constructor
             (e : NIfExp A) -> P e
indNIfExp A P bit var if (NBit b) = bit b
indNIfExp A P bit var if (NVar a) = var a
indNIfExp A P bit var if (NIf a t e) = if a t e
                                          (indNIfExp A P bit var if t)  
                                          (indNIfExp A P bit var if e)

sound_nif  : (A:Set) -> (f:A -> Bool) -> (c,t,e : NIfExp A) ->
              Eq Bool (eval A f (If (emb A c) (emb A t) (emb A e))) 
                      (eval A f (emb A (nif A c t e)))

sound_norm : (A:Set) -> (f:A -> Bool) -> (e:IfExp A) ->
              Eq Bool (eval A f e) 
                      (eval A f (emb A (norm A e)))
